package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the Common Public License
 * 1.0. Applying this annotation will add a Bundle-License clause.
 */
@BundleLicense(name = "http://opensource.org/licenses/cpl1.0.php", link = "http://www.ibm.com/developerworks/library/os-cpl.html", description = "aims of supporting and encouraging collaborative open-source development while still retaining the ability to use the CPL'd content with software licensed under other licenses, including many proprietary licenses.")
public @interface CPL_1_0 {

}
