package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the GNU General Public
 * License 2.0. Applying this annotation will add a Bundle-License clause.
 */

@BundleLicense(name = "http://opensource.org/licenses/GPL-2.0", link = "http://www.gnu.org/licenses/gpl-2.0.html", description = "The GNU General Public License version 2 (GPLv2)")
public @interface GPL_2_0 {}
