package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the Common Development and
 * Distribution license. Applying this annotation will add a Bundle-License
 * clause.
 */

@BundleLicense(name = "http://opensource.org/licenses/CDDL-1.0", link = "http://en.wikipedia.org/wiki/Common_Development_and_Distribution_License", description = "Common Development and Distribution License 1.0")
public @interface CDDL_1_0 {}
