package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the Mozilla Public License
 * 2.0. Applying this annotation will add a Bundle-License clause.
 */
@BundleLicense(name = "http://opensource.org/licenses/MPL-2.0", link = "http://www.mozilla.org/MPL/2.0/", description = "Mozilla Public License 2.0")
public @interface MPL_2_0 {}
