package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the Apache Software
 * License 2.0. Applying this annotation will add a Bundle-License clause.
 */
@BundleLicense(name = "http://www.opensource.org/licenses/apache2.0.php", link = "http://www.apache.org/licenses/LICENSE-2.0.html", description = "Apache Software License 2.0")
public @interface ASL_2_0 {

}
