package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the Eclipsse Public
 * License 1.0. Applying this annotation will add a Bundle-License clause.
 */

@BundleLicense(name = "http://opensource.org/licenses/EPL-1.0", link = "http://www.eclipse.org/legal/epl-v10.html", description = "Eclipse Public License 1.0")
public @interface EPL_1_0 {}
