package aQute.bnd.annotation.licenses;

import aQute.bnd.annotation.headers.BundleLicense;

/**
 * An annotation to indicate that the type depends on the GNU Lesser Public
 * License 2.1. Applying this annotation will add a Bundle-License clause.
 */

@BundleLicense(name = "http://opensource.org/licenses/LGPL-2.1", link = "http://www.gnu.org/licenses/lgpl-2.1.html", description = "GNU Lesser General Public License, version 2.1")
public @interface LGPL_2_1 {

}
